/*
 * Copyright 2005-2010 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wamblee.general;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

/**
 * Thread-specific proxy is used to create implementations of interfaces that
 * delegate to a thread-specific implementation of the service.
 * 
 * It can be used for instance to create a contextual reference to an entity manager 
 * that delegates to a thread-specific instance. 
 * 
 * The {@link #set(Object)} method sets the current service instance for the current thread. 
 * The {@link #get()} method gets the current service instance for the current thread. 
 * The {@link #getProxy()} method gets a proxy that will delegate at runtime to the thread-specific 
 * instance. The result from this method can be passed at construction of an object that will be used
 * by multiple threads. 
 * 
 * This class is mostly used by other test tools. 
 * 
 * @param T
 *            Interface to proxy.
 * @author Erik Brakkee
 * 
 */
public class ThreadSpecificProxyFactory<T> {
    private class ThreadSpecificInvocationHandler implements InvocationHandler {

        @Override
        public Object invoke(Object aProxy, Method aMethod, Object[] aArgs)
            throws Throwable {
            try {
                return aMethod.invoke(svc.get(), aArgs);
            } catch (InvocationTargetException e) {
                throw e.getCause();
            }
        }
    }

    private ThreadLocal<T> svc;
    private Class clazz;
    private T proxy; 

    /**
     * Constructs the factory.
     * 
     * @param aClass
     *            Interface class of the service to proxy.
     */
    public ThreadSpecificProxyFactory(Class<T> aClass) {
        if (!aClass.isInterface()) {
            throw new IllegalArgumentException("Class " + aClass.getName() +
                " is not an interface");
        }
        svc = new ThreadLocal<T>();
        clazz = aClass;
        proxy = createProxy(); 
    }

    /**
     * Sets the thread-specific service.
     * 
     * @param aService
     *            Service, use null value to reset.
     */
    public void set(T aService) {
        svc.set(aService);
    }
    
    /**
     * Gets the current thread-specific service.
     * To get a contextual reference that can be used by any thread but delegates to a thread-specific
     * instance, use {@link #getProxy()}.  
     * @return Service. 
     */
    public T get() { 
        return svc.get();
    }

    /**
     * Gets the proxy that delegates to the thread-specific instance set by
     * {@link #set(Object)}
     * 
     * @return Proxy.
     */
    public T getProxy() {
        return proxy;
    }

    private T createProxy() {
        InvocationHandler handler = new ThreadSpecificInvocationHandler();
        Class proxyClass = Proxy.getProxyClass(clazz.getClassLoader(),
            new Class[] { clazz });
        T proxyObj;
        try {
            proxyObj = (T) proxyClass.getConstructor(
                new Class[] { InvocationHandler.class }).newInstance(
                new Object[] { handler });
            return proxyObj;
        } catch (Exception e) {
            throw new RuntimeException("Could not create proxy for " +
                clazz.getName(), e);
        }
    }
}
